#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import sys


# Global setting to enable, disable, or force the use console codes when the
# print_* functions are used. By default (ON), color codes will only be used if
# the destination is a TTY.
COLORS_OFF, COLORS_ON, COLORS_FORCED = range(3)


def set_color_output(clr):
    """Set use of colors in print functions; possible values are COLORS_OFF,
    COLORS_ON, and COLORS_FORCED.
    """
    global _COLORS
    if clr not in (COLORS_OFF, COLORS_ON, COLORS_FORCED):
        raise ValueError("Invalid value in set_color_output; must be one of "
                         "COLORS_OFF, COLORS_ON, or COLORS_FORCED.")

    _COLORS = clr


def print_msg(*vargs, **kwargs):
    """Equivalent to print. Currently does not apply a color to the text"""
    print(*vargs, **kwargs)


def print_debug(*vargs, **kwargs):
    """Equivalent to print, but prints using shell colorcodes (blue)."""
    _do_print_color(*vargs, colorcode=36, **kwargs)


def print_info(*vargs, **kwargs):
    """Equivalent to print, but prints using shell colorcodes (green)."""
    _do_print_color(*vargs, colorcode=32, **kwargs)


def print_err(*vargs, **kwargs):
    """Equivalent to print, but prints using shell colorcodes (red)."""
    _do_print_color(*vargs, colorcode=31, **kwargs)


def print_warn(*vargs, **kwargs):
    """Equivalent to print, but prints using shell colorcodes (yellow)."""
    _do_print_color(*vargs, colorcode=33, **kwargs)


def print_disabled(*vargs, **kwargs):
    """Equivalent to print, but prints using shell colorcodes (gray)."""
    _do_print_color(*vargs, colorcode=30, **kwargs)


def _do_print_color(*vargs, **kwargs):
    """Utility function: Prints using shell colors."""
    colorcode = kwargs.pop("colorcode")
    destination = kwargs.pop("file", sys.stderr)

    # No colors if output is redirected (e.g. less, file, etc.)
    colors_on = _COLORS != COLORS_OFF
    colors_forced = _COLORS == COLORS_FORCED
    if colors_on and (destination.isatty() or colors_forced):
        vargs = list(vargs)
        for (index, varg) in enumerate(vargs):
            varg_lines = []
            # Newlines terminate the color-code for e.g. 'less', so ensure that
            # each line is color-coded, while preserving the list of arguments
            for line in str(varg).split("\n"):
                varg_lines.append("\033[00;%im%s\033[00m" % (colorcode, line))
            vargs[index] = "\n".join(varg_lines)

    print(*vargs, file=destination, **kwargs)

    if '\n' in kwargs.get('end', '\n'):
        destination.flush()

_COLORS = COLORS_ON



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import re
import random

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import \
     AtomicCmdBuilder, \
     use_customizable_cli_parameters, \
     create_customizable_cli_parameters


RAXML_VERSION = versions.Requirement(call   = ("raxmlHPC", "-version"),
                                     search = r"version (\d+)\.(\d+)\.(\d+)",
                                     checks = versions.GE(7, 3, 2))
RAXML_PTHREADS_VERSION = versions.Requirement(call   = ("raxmlHPC-PTHREADS", "-version"),
                                              search = r"version (\d+)\.(\d+)\.(\d+)",
                                              checks = versions.GE(7, 3, 2))


class RAxMLReduceNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, input_partition, output_alignment, output_partition, dependencies = ()):
        command = AtomicCmdBuilder("raxmlHPC")

        # Read and (in the case of empty columns) reduce input
        command.set_option("-f", "c")
        # Output files are saved with a .Pypeline postfix, and subsequently renamed
        command.set_option("-n", "Pypeline")
        # Model required, but not used
        command.set_option("-m", "GTRGAMMA")
        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")

        # Symlink to sequence and partitions, to prevent the creation of *.reduced files outside temp folder
        # In addition, it may be nessesary to remove the .reduced files if created
        command.set_option("-s", "%(TEMP_IN_ALIGNMENT)s")
        command.set_option("-q", "%(TEMP_IN_PARTITION)s")

        command.set_kwargs(IN_ALIGNMENT      = input_alignment,
                          IN_PARTITION      = input_partition,

                          TEMP_IN_ALIGNMENT = "RAxML_alignment",
                          TEMP_IN_PARTITION = "RAxML_partitions",
                          TEMP_OUT_INFO     = "RAxML_info.Pypeline",

                          OUT_ALIGNMENT     = output_alignment,
                          OUT_PARTITION     = output_partition,
                          CHECK_VERSION     = RAXML_VERSION)

        return {"command" : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._kwargs = parameters.command.kwargs
        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<RAxMLReduce: '%s' -> '%s'>" \
                                     % (parameters.input_alignment, parameters.output_alignment),
                             dependencies = parameters.dependencies)


    def _setup(self, config, temp):
        for key in ("IN_ALIGNMENT", "IN_PARTITION"):
            source      = os.path.abspath(self._kwargs[key])
            destination = os.path.join(temp, self._kwargs["TEMP_" + key])

            os.symlink(source, destination)

        CommandNode._setup(self, config, temp)


    def _teardown(self, config, temp):
        for postfix in ("ALIGNMENT", "PARTITION"):
            filenames = [self._kwargs["TEMP_IN_" + postfix],
                         self._kwargs["TEMP_IN_" + postfix] + ".reduced",
                         self._kwargs["OUT_" + postfix]]

            for (source, destination) in zip(filenames, filenames[1:]):
                source      = fileutils.reroot_path(temp, source)
                destination = fileutils.reroot_path(temp, destination)

                if not os.path.exists(destination):
                    fileutils.copy_file(source, destination)
                os.remove(source)

        CommandNode._teardown(self, config, temp)


class RAxMLBootstrapNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, input_partition, template, start = 0, bootstraps = 50, dependencies = ()):
        command = AtomicCmdBuilder("raxmlHPC", set_cwd = True)

        # Read and (in the case of empty columns) reduce input
        command.set_option("-f", "j")
        # Output files are saved with a .Pypeline postfix, and subsequently renamed
        command.set_option("-n", "Pypeline")
        # Model required, but not used
        command.set_option("-m", "GTRGAMMA")
        # Set random seed for bootstrap generation. May be set to a fixed value to allow replicability.
        command.set_option("-b", int(random.random() * 2**31 - 1), fixed = False)
        # Generate a single bootstrap alignment (makes growing the number of bootstraps easier).
        command.set_option("-N", int(bootstraps), fixed = False)

        # Symlink to sequence and partitions, to prevent the creation of *.reduced files outside temp folder
        # In addition, it may be nessesary to remove the .reduced files if created
        command.set_option("-s", "input.alignment")
        command.set_option("-q", "input.partition")

        bootstrap_files = {"IN_ALIGNMENT" : input_alignment,
                           "IN_PARTITION" : input_partition,
                           "TEMP_OUT_INF" : "RAxML_info.Pypeline",
                           "TEMP_OUT_ALN" : "input.alignment",
                           "TEMP_OUT_PAR" : "input.partition",
                           "CHECK_VERSION": RAXML_VERSION}

        for (index, (_, filename)) in enumerate(cls._bootstraps(template, bootstraps, start)):
            bootstrap_files["OUT_BS_%03i" % index] = filename
        command.set_kwargs(**bootstrap_files)

        return {"command" : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._input_alignment = parameters.input_alignment
        self._input_partition = parameters.input_partition
        self._output_template = parameters.template
        self._bootstrap_num   = parameters.bootstraps
        self._bootstrap_start = parameters.start

        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<RAxMLBootstrap: '%s' -> '%s' (%i .. %i>" \
                                     % (parameters.input_alignment, parameters.template,
                                        parameters.start, parameters.start + parameters.bootstraps - 1),
                             dependencies = parameters.dependencies)

    def _setup(self, config, temp):
        os.symlink(os.path.realpath(self._input_alignment), os.path.join(temp, "input.alignment"))
        os.symlink(os.path.realpath(self._input_partition), os.path.join(temp, "input.partition"))


    def _teardown(self, config, temp):
        template   = self._output_template
        bootstraps = self._bootstrap_num
        start      = self._bootstrap_start
        for (src_file, dst_file) in self._bootstraps(template, bootstraps, start):
            src_file = os.path.join(temp, src_file)
            dst_file = fileutils.reroot_path(temp, dst_file)
            fileutils.move_file(src_file, dst_file)
        CommandNode._teardown(self, config, temp)

    @classmethod
    def _bootstraps(cls, template, number, start):
        for bootstrap in range(number):
            src_file = "input.alignment.BS%i" % (bootstrap,)
            dst_file = template % (bootstrap + start,)
            yield (src_file, dst_file)


class RAxMLRapidBSNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, output_template, input_partition=None,
                  threads=1, dependencies=()):
        """
        Arguments:
        input_alignment  -- An alignment file in a format readable by RAxML.
        input_partition  -- A set of partitions in a format readable by RAxML.
        output_template  -- A template string used to construct final filenames. Should consist
                            of a full path, including a single '%s', which is replaced with the
                            variable part of RAxML output files (e.g. 'info', 'bestTree', ...).
                            Example destination: '/disk/project/SN013420.RAxML.%s'
                            Example output:      '/disk/project/SN013420.RAxML.bestTree'
        """

        if threads > 1:
            command = AtomicCmdBuilder("raxmlHPC-PTHREADS")
            command.set_option("-T", threads)
            version = RAXML_PTHREADS_VERSION
        else:
            command = AtomicCmdBuilder("raxmlHPC")
            version = RAXML_VERSION

        # Perform rapid bootstrapping
        command.set_option("-f", "a")
        # Output files are saved with a .PALEOMIX postfix, and subsequently renamed
        command.set_option("-n", "PALEOMIX")
        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")
        # Symlink to sequence and partitions, to prevent the creation of *.reduced files outside temp folder
        # In addition, it may be nessesary to remove the .reduced files if created
        command.set_option("-s", "%(TEMP_OUT_ALN)s")

        if input_partition is not None:
            command.set_option("-q", "%(TEMP_OUT_PART)s")
            command.set_kwargs(IN_PARTITION=input_partition,
                               TEMP_OUT_PART=os.path.basename(input_partition),
                               TEMP_OUT_PART_R=os.path.basename(input_partition) + ".reduced")

        command.set_kwargs(  # Auto-delete: Symlinks and .reduced files that RAxML may generate
                           TEMP_OUT_ALN=os.path.basename(input_alignment),
                           TEMP_OUT_ALN_R=os.path.basename(input_alignment) + ".reduced",

                           # Input files, are not used directly (see below)
                           IN_ALIGNMENT=input_alignment,

                           # Final output files, are not created directly
                           OUT_INFO=output_template % "info",
                           OUT_BESTTREE=output_template % "bestTree",
                           OUT_BOOTSTRAP=output_template % "bootstrap",
                           OUT_BIPART=output_template % "bipartitions",
                           OUT_BIPARTLABEL=output_template % "bipartitionsBranchLabels",

                           CHECK_VERSION=version)

        # Use the GTRGAMMA model of NT substitution by default
        command.set_option("-m", "GTRGAMMAI", fixed=False)
        # Enable Rapid Boostrapping and set random seed. May be set to a fixed value to allow replicability.
        command.set_option("-x", int(random.random() * 2**31 - 1), fixed=False)
        # Set random seed for parsimony inference. May be set to a fixed value to allow replicability.
        command.set_option("-p", int(random.random() * 2**31 - 1), fixed=False)
        # Terminate bootstrapping upon convergence, rather than after a fixed number of repetitions
        command.set_option("-N", "autoMRE", fixed=False)

        return {"command": command}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._symlinks = [parameters.input_alignment,
                          parameters.input_partition]
        self._template = os.path.basename(parameters.output_template)

        CommandNode.__init__(self,
                             command=parameters.command.finalize(),
                             description="<RAxMLRapidBS: '%s' -> '%s'>"
                             % (parameters.input_alignment,
                                parameters.output_template % ("*",)),
                             threads=parameters.threads,
                             dependencies=parameters.dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        # Required to avoid the creation of files outside the temp folder
        for filename in self._symlinks:
            if filename is not None:
                source = os.path.abspath(filename)
                destination = os.path.join(temp, os.path.basename(filename))

                os.symlink(source, destination)

    def _teardown(self, config, temp):
        for filename in os.listdir(temp):
            match = re.match("RAxML_(.*).PALEOMIX", filename)
            if match:
                source = os.path.join(temp, filename)
                destination = os.path.join(temp, self._template % match.groups())

                fileutils.move_file(source, destination)

        CommandNode._teardown(self, config, temp)


class RAxMLParsimonyTreeNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_alignment, input_partitions, output_tree, dependencies = ()):
        command = AtomicCmdBuilder("raxmlHPC")

        # Compute a randomized parsimony starting tree
        command.set_option("-y")
        # Output files are saved with a .Pypeline postfix, and subsequently renamed
        command.set_option("-n", "Pypeline")
        # Model required, but not used
        command.set_option("-m", "GTRGAMMA")
        # Ensures that output is saved to the temporary directory
        command.set_option("-w", "%(TEMP_DIR)s")
        # Set random seed for bootstrap generation. May be set to a fixed value to allow replicability.
        command.set_option("-p", int(random.random() * 2**31 - 1), fixed = False)

        # Symlink to sequence and partitions, to prevent the creation of *.reduced files outside temp folder
        command.set_option("-s", "%(TEMP_OUT_ALIGNMENT)s")
        command.set_option("-q", "%(TEMP_OUT_PARTITION)s")

        command.set_kwargs(IN_ALIGNMENT       = input_alignment,
                           IN_PARTITION       = input_partitions,

                           # TEMP_OUT_ is used to automatically remove these files
                           TEMP_OUT_ALIGNMENT = "RAxML_alignment",
                           TEMP_OUT_PARTITION = "RAxML_partitions",
                           TEMP_OUT_INFO      = "RAxML_info.Pypeline",

                           OUT_TREE           = output_tree,

                           CHECK_VERSION      = RAXML_VERSION)

        return {"command" : command}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._input_alignment  = parameters.input_alignment
        self._input_partitions = parameters.input_partitions
        self._output_tree      = parameters.output_tree

        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = "<RAxMLParsimonyTree: '%s' -> '%s'>" \
                                     % (parameters.input_alignment, parameters.output_tree),
                             dependencies = parameters.dependencies)


    def _setup(self, config, temp):
        os.symlink(os.path.abspath(self._input_alignment),  os.path.join(temp, "RAxML_alignment"))
        os.symlink(os.path.abspath(self._input_partitions), os.path.join(temp, "RAxML_partitions"))
        CommandNode._setup(self, config, temp)


    def _teardown(self, config, temp):
        basename = os.path.basename(self._output_tree)
        os.rename(os.path.join(temp, "RAxML_parsimonyTree.Pypeline"),
                  os.path.join(temp, basename))

        CommandNode._teardown(self, config, temp)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import \
    CommandNode
from paleomix.atomiccmd.sets import \
    ParallelCmds
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions
import paleomix.tools.factory as factory


_VERSION_CHECK = versions.Requirement(call=("AdapterRemoval", "--version"),
                                      search=r"ver. (\d+)\.(\d+)\.(\d+)",
                                      checks=versions.GE(2, 1, 5))


class SE_AdapterRemovalNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_files, output_prefix, output_format="bz2",
                  threads=1, dependencies=()):
        # See below for parameters in common between SE/PE
        cmd = _get_common_parameters(output_format, threads=threads)

        # Prefix for output files, ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s." + output_format
        cmd.set_kwargs(TEMP_OUT_BASENAME=os.path.basename(output_prefix),

                       OUT_SETTINGS=output_prefix + ".settings",
                       OUT_MATE_1=output_tmpl % ("truncated",),
                       OUT_DISCARDED=output_tmpl % ("discarded",))

        if len(input_files) > 1:
            # Uncompressed reads (piped from 'paleomix cat')
            cmd.set_option("--file1", "%(TEMP_IN_READS_1)s")
            cmd.set_kwargs(TEMP_IN_READS_1="uncompressed_input")
        else:
            cmd.set_option("--file1", "%(IN_READS_1)s")
            cmd.set_kwargs(IN_READS_1=input_files[0])

        return {"command": cmd,
                "threads": threads,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()

        self._multi_file_input = len(parameters.input_files) > 1
        if self._multi_file_input:
            cat = _build_cat_command(parameters.input_files, "uncompressed_input")
            command = ParallelCmds((command, cat))

        CommandNode.__init__(self,
                             command=command,
                             threads=parameters.threads,
                             description="<AdapterRM (SE): %s -> '%s.*'>"
                             % (fileutils.describe_files(parameters.input_files),
                                parameters.output_prefix),
                             dependencies=parameters.dependencies)

    def _setup(self, config, temp):
        if self._multi_file_input:
            os.mkfifo(os.path.join(os.path.join(temp, "uncompressed_input")))

        CommandNode._setup(self, config, temp)


class PE_AdapterRemovalNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_files_1, input_files_2, output_prefix,
                  output_format="bz2", collapse=True, threads=1,
                  dependencies=()):
        if len(input_files_1) != len(input_files_2):
            raise ValueError("Unequal number of mate 1 and mate 2 files")

        cmd = _get_common_parameters(output_format, threads=threads)

        # Prefix for output files, to ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s." + output_format
        cmd.set_kwargs(TEMP_OUT_BASENAME=os.path.basename(output_prefix),

                       OUT_SETTINGS=output_prefix + ".settings",
                       OUT_READS_1=output_tmpl % ("pair1.truncated",),
                       OUT_READS_2=output_tmpl % ("pair2.truncated",),
                       OUT_SINGLETON=output_tmpl % ("singleton.truncated",),
                       OUT_DISCARDED=output_tmpl % ("discarded",))

        if collapse:
            cmd.set_option("--collapse")

            cmd.set_kwargs(OUT_COLLAPSED=output_tmpl % ("collapsed",),
                           OUT_COLLAPSED_TRUNC=output_tmpl
                           % ("collapsed.truncated",))

        if len(input_files_1) > 1:
            # Uncompressed reads (piped from 'paleomix cat')
            cmd.set_option("--file1", "%(TEMP_IN_READS_1)s")
            cmd.set_option("--file2", "%(TEMP_IN_READS_2)s")
            cmd.set_kwargs(TEMP_IN_READS_1="uncompressed_input_1",
                           TEMP_IN_READS_2="uncompressed_input_2")
        else:
            cmd.set_option("--file1", "%(IN_READS_1)s")
            cmd.set_option("--file2", "%(IN_READS_2)s")
            cmd.set_kwargs(IN_READS_1=input_files_1[0],
                           IN_READS_2=input_files_2[0])

        return {"command": cmd,
                "threads": threads,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()
        self._multi_file_input = len(parameters.input_files_1) > 1
        if self._multi_file_input:
            cat_1 = _build_cat_command(parameters.input_files_1, "uncompressed_input_1")
            cat_2 = _build_cat_command(parameters.input_files_2, "uncompressed_input_2")
            command = ParallelCmds((command, cat_1, cat_2))

        CommandNode.__init__(self,
                             command=command,
                             threads=parameters.threads,
                             description="<AdapterRM (PE): %s -> '%s.*'>"
                             % (fileutils.describe_paired_files(parameters.input_files_1,
                                                                parameters.input_files_2),
                                parameters.output_prefix),
                             dependencies=parameters.dependencies)

    def _setup(self, config, temp):
        if self._multi_file_input:
            os.mkfifo(os.path.join(os.path.join(temp, "uncompressed_input_1")))
            os.mkfifo(os.path.join(os.path.join(temp, "uncompressed_input_2")))

        CommandNode._setup(self, config, temp)


def _build_cat_command(input_files, output_file):
    cat = factory.new("cat")
    cat.set_option("--output", "%(TEMP_OUT_CAT)s")
    cat.set_kwargs(TEMP_OUT_CAT=output_file)
    cat.add_multiple_values(input_files)

    return cat.finalize()


def _get_common_parameters(output_format, threads=1):
    cmd = AtomicCmdBuilder("AdapterRemoval",
                           CHECK_VERSION=_VERSION_CHECK)

    if output_format == "bz2":
        cmd.set_option("--bzip2")
    elif output_format == "gz":
        cmd.set_option("--gzip")
    else:
        raise ValueError("Invalid output compression %r" % (output_format,))

    # Trim Ns at read ends
    cmd.set_option("--trimns", fixed=False)
    # Trim low quality scores
    cmd.set_option("--trimqualities", fixed=False)

    # Fix number of threads to ensure consistency when scheduling node
    cmd.set_option("--threads", threads)

    return cmd



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import re

from paleomix.node import CommandNode, NodeError
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import SequentialCmds
from paleomix.common.formats.fasta import FASTA
from paleomix.common.utilities import safe_coerce_to_frozenset

import paleomix.common.fileutils as fileutils


class CodemlNode(CommandNode):
    def __init__(self, control_file, sequence_file, trees_file, output_tar,
                 exclude_groups=(), dependencies=()):
        self._exclude_groups = safe_coerce_to_frozenset(exclude_groups)
        self._control_file = control_file
        self._sequence_file = sequence_file
        self._trees_file = trees_file

        paml_cmd = AtomicCmd(["codeml", "template.ctl"],
                             IN_CONTROL_FILE  = control_file,
                             IN_SEQUENCE_FILE = sequence_file,
                             IN_TREES_FILE    = trees_file,
                             TEMP_OUT_CTL     = "template.ctl",
                             TEMP_OUT_SEQS    = "template.seqs",
                             TEMP_OUT_TREES   = "template.trees",
                             TEMP_OUT_STDOUT  = "template.stdout",
                             TEMP_OUT_STDERR  = "template.stderr",
                             TEMP_OUT_4FOLD   = "4fold.nuc",
                             IN_STDIN         = "/dev/null", # Prevent promts from blocking
                             set_cwd          = True,
                             **CodemlNode._get_codeml_files("TEMP_OUT_CODEML"))

        tar_pairs = CodemlNode._get_codeml_files("TEMP_IN_CODEML")
        tar_files = ["%%(%s)s" % (key,) for key in tar_pairs]
        tar_cmd  = AtomicCmd(["tar", "cvzf", "%(OUT_FILE)s"] + tar_files,
                             OUT_FILE = output_tar,
                             set_cwd  = True,
                             **tar_pairs)

        CommandNode.__init__(self,
                             description  = "<CodemlNode: %r -> %r>" % (sequence_file, output_tar),
                             command      = SequentialCmds([paml_cmd, tar_cmd]),
                             dependencies = dependencies)

    def _setup(self, _config, temp):
        self._update_ctl_file(source      = self._control_file,
                              destination = os.path.join(temp, "template.ctl"))

        os.symlink(os.path.abspath(self._trees_file), os.path.join(temp, "template.trees"))
        with open(os.path.join(temp, "template.seqs"), "w") as handle:
            for record in FASTA.from_file(self._sequence_file):
                if record.name not in self._exclude_groups:
                    name = record.name
                    sequence = record.sequence.upper()
                    handle.write("%s\n" % (FASTA(name, None, sequence),))

    def _run(self, config, temp):
        try:
            CommandNode._run(self, config, temp)
        except NodeError, error:
            if self._command.join() == [1, None]:
                with open(fileutils.reroot_path(temp, "template.stdout")) as handle:
                    lines = handle.readlines()
                if lines and ("Giving up." in lines[-1]):
                    error = NodeError("%s\n\n%s" % (error, lines[-1]))
            raise error

    @classmethod
    def _update_ctl_file(cls, source, destination):
        with open(source) as handle:
            template = handle.read()

        # TODO: Do check before running everything!
        template, count = re.subn(r'(\bseqfile\s*=).*',  r'\1 template.seqs', template)
        assert count == 1, count
        template, count = re.subn(r'(\btreefile\s*=).*', r'\1 template.trees', template)
        assert count == 1, count
        template, count = re.subn(r'(\boutfile\s*=).*',  r'\1 mlc',       template)
        assert count == 1, count

        with open(destination, "w") as handle:
            handle.write(template)

    @classmethod
    def _get_codeml_files(cls, key_type):
        results = {}
        codeml_files = ["mlc", "2NG.dN", "2NG.dS", "2NG.t",
                        "lnf", "rst", "rst1", "rub"]

        for filename in codeml_files:
            key = "%s_%s" % (key_type, filename.upper().replace(".", "_"))
            results[key] = filename
        return results


def build_codeml_nodes(options, settings, regions, filtering, dependencies):
    in_postfix, out_postfix, afa_ext = "", "", ".afa"
    if any(filtering.itervalues()):
        in_postfix = out_postfix = ".filtered"
    if not settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]:
        out_postfix = ".unaligned" + out_postfix
        afa_ext = ".fasta"

    codeml = settings["PAML"]["codeml"]
    subset_key = codeml["SubsetRegions"].get(regions["Name"])
    sequences = regions["Sequences"][subset_key]
    sequencedir = os.path.join(options.destination, "alignments", regions["Name"] + in_postfix)
    destination = os.path.join(options.destination, "paml", "codeml", regions["Name"] + out_postfix)

    fasta_files = {}
    for sequence in sequences:
        fasta_files[sequence] = os.path.join(sequencedir, sequence + afa_ext)

    codeml_nodes = []
    for (ctl_name, ctl_files) in codeml.iteritems():
        # This dictionary also contains the "ExcludeSamples" option
        if ctl_name in ("ExcludeSamples", "SubsetRegions"):
            continue

        for (sequence, filename) in fasta_files.iteritems():
            output_tar = os.path.join(destination, "%s.%s.tar.gz" % (sequence, ctl_name))
            ctl_file = ctl_files["ControlFile"].format(Name=sequence)
            tree_file = ctl_files["TreeFile"].format(Name=sequence)

            node = CodemlNode(control_file=ctl_file,
                              trees_file=tree_file,
                              sequence_file=filename,
                              output_tar=output_tar,
                              exclude_groups=codeml["ExcludeSamples"],
                              dependencies=dependencies)
            codeml_nodes.append(node)

    return codeml_nodes


def chain_codeml(_pipeline, options, makefiles):
    destination = options.destination  # Move to makefile
    for makefile in makefiles:
        nodes = []
        filtering = makefile["Project"]["FilterSingletons"]
        options.destination = os.path.join(destination, makefile["Project"]["Title"])

        for regions in makefile["Project"]["Regions"].itervalues():
            nodes.extend(build_codeml_nodes(options, makefile, regions, filtering, makefile["Nodes"]))
        makefile["Nodes"] = tuple(nodes)
    options.destination = destination



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import types
import traceback
import collections

import paleomix.common.fileutils as fileutils
from paleomix.common.utilities import \
     safe_coerce_to_frozenset

from paleomix.atomiccmd.command import \
    CmdError


class NodeError(RuntimeError):
    pass


class CmdNodeError(NodeError):
    pass


class NodeUnhandledException(NodeError):
    """This exception is thrown by Node.run() if a non-NodeError exception
    is raised in a subfunction (e.g. _setup, _run, or _teardown). The text
    for this exception will include both the original error message and a
    stacktrace for that error."""
    pass


class Node(object):
    def __init__(self, description = None, threads = 1,
                 input_files = (), output_files = (),
                 executables = (), auxiliary_files = (),
                 requirements = (), dependencies = ()):

        if not isinstance(description, _DESC_TYPES):
            raise TypeError("'description' must be None or a string, not %r" \
                            % (description.__class__.__name__,))

        self.__description   = description
        self.input_files     = self._validate_files(input_files)
        self.output_files    = self._validate_files(output_files)
        self.executables     = self._validate_files(executables)
        self.auxiliary_files = self._validate_files(auxiliary_files)
        self.requirements    = self._validate_requirements(requirements)

        self.threads         = self._validate_nthreads(threads)
        self.dependencies    = self._collect_nodes(dependencies)

        # If there are no input files, the node cannot be re-run based on
        # changes to the input, and nodes with output but no input are not
        # expected based on current usage.
        if not self.input_files and self.output_files:
            raise NodeError("Node not dependant upon input files: %s" % self)

    def run(self, config):
        """Runs the node, by calling _setup, _run, and _teardown in that order.
        Prior to calling these functions, a temporary dir is created using the
        'temp_root' prefix from the config object. Both the config object and
        the temporary dir are passed to the above functions. The temporary
        dir is removed after _teardown is called, and all expected files
        should have been removed/renamed at that point.

        Any non-NodeError exception raised in this function is wrapped in a
        NodeUnhandledException, which includes a full backtrace. This is needed
        to allow showing these in the main process."""

        try:
            temp = None
            temp = self._create_temp_dir(config)

            self._setup(config, temp)
            self._run(config, temp)
            self._teardown(config, temp)
            self._remove_temp_dir(temp)
        except NodeError, error:
            self._write_error_log(temp, error)
            raise NodeError("Error(s) running Node:\n\tTemporary directory: %s\n\n%s" \
                            % (repr(temp), error))
        except Exception, error:
            self._write_error_log(temp, error)
            raise NodeUnhandledException("Error(s) running Node:\n\tTemporary directory: %s\n\n%s" \
                                         % (repr(temp), traceback.format_exc()))

    def _create_temp_dir(self, config):
        """Called by 'run' in order to create a temporary folder.

        'config' is expected to have a property .temp_root under
        which the temporary folder is created."""
        return fileutils.create_temp_dir(config.temp_root)

    def _remove_temp_dir(self, temp):
        """Called by 'run' in order to remove an (now) empty temporary folder."""
        os.rmdir(temp)

    def _setup(self, _config, _temp):
        """Is called prior to '_run()' by 'run()'. Any code used to copy/link files,
        or other steps needed to ready the node for running may be carried out in this
        function. Checks that required input files exist, and raises an NodeError if
        this is not the case."""
        if fileutils.missing_executables(self.executables):
            raise NodeError("Executable(s) does not exist for node: %s" % (self,))
        self._check_for_missing_files(self.input_files, "input")
        self._check_for_missing_files(self.auxiliary_files, "auxiliary")

    def _run(self, _config, _temp):
        pass

    def _teardown(self, _config, _temp):
        self._check_for_missing_files(self.output_files, "output")

    def __str__(self):
        """Returns the description passed to the constructor, or a default
        description if no description was passed to the constructor."""
        if self.__description:
            return self.__description
        return repr(self)

    def __getstate__(self):
        """Called by pickle/cPickle to determine what to pickle; this is
        overridden to avoid pickling of requirements, dependencies, which would
        otherwise greatly inflate the amount of information that needs to be
        pickled."""
        obj_dict = self.__dict__.copy()
        obj_dict["requirements"] = None
        obj_dict["dependencies"] = None
        return obj_dict

    def _write_error_log(self, temp, error):
        if not (temp and os.path.isdir(temp)):
            return

        prefix = "\n                   "
        message = ["Command          = %s" % (" ".join(sys.argv),),
                   "CWD              = %s" % (os.getcwd(),),
                   "Node             = %s" % (str(self),),
                   "Threads          = %i" % (self.threads,),
                   "Input files      = %s" % (prefix.join(sorted(self.input_files)),),
                   "Output files     = %s" % (prefix.join(sorted(self.output_files)),),
                   "Auxiliary files  = %s" % (prefix.join(sorted(self.auxiliary_files)),),
                   "Executables      = %s" % (prefix.join(sorted(self.executables)),),
                   "",
                   "Errors =\n%s\n" % (error,)]
        message = "\n".join(message)

        try:
            with open(os.path.join(temp, "pipe.errors"), "w") as handle:
                handle.write(message)
        except OSError, oserror:
            sys.stderr.write("ERROR: Could not write failure log: %s\n" % (oserror,))

    def _collect_nodes(self, nodes):
        if nodes is None:
            return frozenset()

        nodes = safe_coerce_to_frozenset(nodes)
        bad_nodes = [node for node in nodes if not isinstance(node, Node)]

        if bad_nodes:
            bad_nodes = [repr(node) for node in bad_nodes]
            message = "Dependency-list contain non-Node objects:\n" \
                "\t- Command: %s\n\t- Objects: %s" \
                % (self, "\n\t           ".join(bad_nodes))
            raise TypeError(message)

        return nodes

    def _check_for_missing_files(self, filenames, description):
        missing_files = fileutils.missing_files(filenames)
        if missing_files:
            message = "Missing %s files for command:\n\t- Command: %s\n\t- Files: %s" \
                % (description, self, "\n\t         ".join(missing_files))
            raise NodeError(message)

    @classmethod
    def _validate_requirements(cls, requirements):
        requirements = safe_coerce_to_frozenset(requirements)
        for requirement in requirements:
            if not isinstance(requirement, collections.Callable):
                raise TypeError("'requirements' must be callable, not %r"
                                % (type(requirement),))
        return requirements

    @classmethod
    def _validate_files(cls, files):
        files = safe_coerce_to_frozenset(files)
        for filename in files:
            if not isinstance(filename, types.StringTypes):
                raise TypeError('Files must be strings, not %r' % filename.__class__.__name__)
        return files

    @classmethod
    def _validate_nthreads(cls, threads):
        if not isinstance(threads, (types.IntType, types.LongType)):
            raise TypeError("'threads' must be a positive integer, not %s" % (type(threads),))
        elif threads < 1:
            raise ValueError("'threads' must be a positive integer, not %i" % (threads,))
        return int(threads)


class CommandNode(Node):
    def __init__(self, command, description=None, threads=1,
                 dependencies=()):
        Node.__init__(self,
                      description  = description,
                      input_files  = command.input_files,
                      output_files = command.output_files,
                      auxiliary_files = command.auxiliary_files,
                      executables  = command.executables,
                      requirements = command.requirements,
                      threads      = threads,
                      dependencies = dependencies)

        self._command = command

    def _run(self, _config, temp):
        """Runs the command object provided in the constructor, and waits for it to
        terminate. If any errors during the running of the command, this function
        raises a NodeError detailing the returned error-codes."""
        try:
            self._command.run(temp)
        except CmdError, error:
            desc = "\n\t".join(str(self._command).split("\n"))
            raise CmdNodeError("%s\n\n%s" % (desc, error))

        return_codes = self._command.join()
        if any(return_codes):
            desc = "\n\t".join(str(self._command).split("\n"))
            raise CmdNodeError(desc)


    def _teardown(self, config, temp):
        required_files = self._command.expected_temp_files
        optional_files = self._command.optional_temp_files
        current_files  = set(os.listdir(temp))

        missing_files = (required_files - current_files)
        if missing_files:
            raise CmdNodeError(("Error running Node, required files not created:\n"
                               "Temporary directory: %r\n"
                               "\tRequired files missing from temporary directory:\n\t    - %s") \
                               % (temp, "\n\t    - ".join(sorted(map(repr, missing_files)))))

        extra_files = current_files - (required_files | optional_files)
        if extra_files:
            raise CmdNodeError("Error running Node, unexpected files created:\n"
                               "\tTemporary directory: %r\n"
                               "\tUnexpected files found in temporary directory:\n\t    - %s" \
                               % (temp, "\n\t    - ".join(sorted(map(repr, extra_files)))))

        self._command.commit(temp)

        Node._teardown(self, config, temp)


# Types that are allowed for the 'description' property
_DESC_TYPES = types.StringTypes + (types.NoneType,)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.atomiccmd.builder import apply_options
from paleomix.nodes.adapterremoval import \
    SE_AdapterRemovalNode, \
    PE_AdapterRemovalNode
from paleomix.nodes.validation import \
    ValidateFASTQFilesNode


class Reads(object):
    def __init__(self, config, record, quality_offset):
        self.quality_offset = quality_offset
        self.files = {}
        self.stats = None
        self.nodes = ()

        tags = record["Tags"]
        self.folder = os.path.join(config.destination, tags["Target"], "reads",
                                   tags["SM"], tags["LB"], tags["PU_cur"])

        lane_type = record.get("Type")
        if lane_type == "Raw":
            self._init_raw_reads(config, record)
        elif lane_type == "Trimmed":
            self._init_pretrimmed_reads(record)
        else:
            assert False, "Unexpected data type in Reads(): %s" \
                % (repr(lane_type))

        for name, value in record["Options"]["ExcludeReads"].iteritems():
            if value:
                self.files.pop(name, None)

    def _init_pretrimmed_reads(self, record):
        self.files.update(record["Data"])
        output_file = os.path.join(self.folder, "reads.pretrimmed.validated")
        input_files = set()
        for (read_type, filename) in self.files.iteritems():
            if read_type == "Paired":
                input_files.add(filename.format(Pair=1))
                input_files.add(filename.format(Pair=2))
            else:
                input_files.add(filename)

        node = ValidateFASTQFilesNode(input_files=input_files,
                                      output_file=output_file,
                                      offset=self.quality_offset)
        self.nodes = (node,)

    def _init_raw_reads(self, config, record):
        ar_options = dict(record["Options"]["AdapterRemoval"])
        # Setup of "--collapsed" is handled by the node itself
        collapse_reads = ar_options.pop("--collapse")
        collapse_reads = collapse_reads or collapse_reads is None

        init_args = {"output_prefix": os.path.join(self.folder, "reads"),
                     "output_format": record["Options"]["CompressionFormat"],
                     "threads": config.adapterremoval_max_threads}
        output_tmpl = "{output_prefix}.%s.{output_format}".format(**init_args)

        if ("SE" in record["Data"]):
            self.files["Single"] = output_tmpl % ("truncated",)
            init_args["input_files"] = record["Data"]["SE"]
            command = SE_AdapterRemovalNode.customize(**init_args)
        else:
            self.files["Singleton"] = output_tmpl % ("singleton.truncated",)
            self.files["Paired"] = output_tmpl % ("pair{Pair}.truncated",)

            if collapse_reads:
                self.files["Collapsed"] = output_tmpl % ("collapsed",)
                self.files["CollapsedTruncated"] = output_tmpl % ("collapsed.truncated",)

            init_args["collapse"] = collapse_reads
            init_args["input_files_1"] = record["Data"]["PE_1"]
            init_args["input_files_2"] = record["Data"]["PE_2"]
            command = PE_AdapterRemovalNode.customize(**init_args)

        # Ensure that any user-specified list of adapters is tracked
        if "--adapter-list" in ar_options:
            adapter_list = ar_options.pop("--adapter-list")
            command.command.set_option("--adapter-list", "%(IN_ADAPTER_LIST)s")
            command.command.set_kwargs(IN_ADAPTER_LIST=adapter_list)

        apply_options(command.command, ar_options)

        output_quality = self.quality_offset
        if output_quality == "Solexa":
            output_quality = "64"

        command.command.set_option("--qualitybase", self.quality_offset)
        command.command.set_option("--qualitybase-output", output_quality)

        self.stats = os.path.join(self.folder, "reads.settings")
        self.nodes = (command.build_node(),)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode, NodeError
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.builder import \
     AtomicCmdBuilder, \
     use_customizable_cli_parameters, \
     create_customizable_cli_parameters
from paleomix.atomiccmd.sets import \
     ParallelCmds
from paleomix.nodes.bwa import \
     _get_node_description, \
     _process_output, \
     _get_max_threads

import paleomix.common.versions as versions



BOWTIE2_VERSION = versions.Requirement(call   = ("bowtie2", "--version"),
                                       search = r"version (\d+)\.(\d+)\.(\d+)",
                                       checks = versions.GE(2, 1, 0))



class Bowtie2IndexNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file, prefix = None, dependencies = ()):
        prefix = prefix if prefix else input_file
        params = _bowtie2_template(("bowtie2-build"), prefix, iotype = "OUT",
                                   IN_FILE = input_file,
                                   TEMP_OUT_PREFIX = os.path.basename(prefix),
                                   CHECK_VERSION = BOWTIE2_VERSION)
        params.add_value("%(IN_FILE)s")
        # Destination prefix, in temp folder
        params.add_value("%(TEMP_OUT_PREFIX)s")

        return {"prefix"       :  prefix,
                "command"      : params,
                "dependencies" : dependencies}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()
        description =  "<Bowtie2 Index '%s' -> '%s.*'>" % (parameters.input_file,
                                                       parameters.prefix)
        CommandNode.__init__(self,
                             command      = command,
                             description  = description,
                             dependencies = parameters.dependencies)



class Bowtie2Node(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file_1, input_file_2, output_file, reference, prefix, threads = 2, dependencies = ()):
        aln = _bowtie2_template(("bowtie2",), prefix,
                                IN_FILE_1  = input_file_1,
        # Setting IN_FILE_2 to None makes AtomicCmd ignore this key
                                IN_FILE_2  = input_file_2 or None,
                                OUT_STDOUT = AtomicCmd.PIPE,
                                CHECK_VERSION = BOWTIE2_VERSION)
        aln.set_option("-x", prefix)

        if input_file_1 and not input_file_2:
            aln.set_option("-U", "%(IN_FILE_1)s")
        elif input_file_1 and input_file_2:
            aln.set_option("-1", "%(IN_FILE_1)s")
            aln.set_option("-2", "%(IN_FILE_2)s")
        else:
            raise NodeError("Input 1, OR both input 1 and input 2 must be specified for Bowtie2 node")

        max_threads = _get_max_threads(reference, threads)
        aln.set_option("--threads", max_threads)

        order, commands = _process_output(aln, output_file, reference, run_fixmate = (input_file_1 and input_file_2))
        commands["aln"] = aln

        return {"commands"     : commands,
                "order"        : ["aln"] + order,
                "threads"      : max_threads,
                "dependencies" : dependencies}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = ParallelCmds([parameters.commands[key].finalize() for key in parameters.order])

        algorithm    = "PE" if parameters.input_file_2 else "SE"
        description  = _get_node_description(name          = "Bowtie2",
                                             algorithm     = algorithm,
                                             input_files_1 = parameters.input_file_1,
                                             input_files_2 = parameters.input_file_2,
                                             prefix        = parameters.prefix,
                                             threads       = parameters.threads)

        CommandNode.__init__(self,
                             command      = command,
                             description  = description,
                             threads      = parameters.threads,
                             dependencies = parameters.dependencies)


def _bowtie2_template(call, prefix, iotype = "IN", **kwargs):
    params = AtomicCmdBuilder(call, **kwargs)
    for postfix in ("1.bt2", "2.bt2", "3.bt2", "4.bt2", "rev.1.bt2", "rev.2.bt2"):
        key = "%s_PREFIX_%s" % (iotype, postfix.upper())
        params.set_kwargs(**{key : (prefix + "." + postfix)})

    return params



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import collections

from paleomix.node import Node
from paleomix.common.fileutils import move_file, reroot_path
from paleomix.common.formats.msa import MSA
from paleomix.common.formats.phylip import interleaved_phy, sequential_phy

from paleomix.common.utilities import \
     safe_coerce_to_frozenset, \
     safe_coerce_to_tuple



_VALID_KEYS = frozenset(["partitions", "filenames"])


class FastaToPartitionedInterleavedPhyNode(Node):
    def __init__(self, infiles, out_prefix, exclude_groups=(), reduce=False,
                 dependencies=(), file_dependencies=()):
        """
        infiles = {names : {"partitions" : ..., "filenames" : [...]}}
        """
        if not (isinstance(infiles, dict)
                and all(isinstance(dd, dict) for dd in infiles.values())):
            raise TypeError("'infiles' must be a dictionary of dictionaries")

        input_filenames = []
        for (name, subdd) in infiles.iteritems():
            if set(subdd) - _VALID_KEYS:
                raise ValueError("Invalid keys found for %r: %s"
                                 % (name, ", ".join(set(subdd) - _VALID_KEYS)))
            elif not isinstance(subdd["filenames"], list):
                raise ValueError("filenames must be a list of strings")
            input_filenames.extend(subdd["filenames"])
        # Optional file dependencies; used to depend on the list of sequcences
        input_filenames.extend(safe_coerce_to_tuple(file_dependencies))

        self._reduce = bool(reduce)
        self._infiles = copy.deepcopy(infiles)
        self._out_prefix = out_prefix
        self._excluded = safe_coerce_to_frozenset(exclude_groups)

        description = "<FastaToPartitionedPhy%s: %i file(s) -> '%s.*'>" % \
            (" (reducing)" if reduce else "", len(infiles), out_prefix)

        Node.__init__(self,
                      description=description,
                      input_files=input_filenames,
                      output_files=[out_prefix + ".phy",
                                    out_prefix + ".partitions"],
                      dependencies=dependencies)

    def _run(self, _config, temp):
        merged_msas = []
        for (name, files_dd) in sorted(self._infiles.iteritems()):
            partitions = files_dd["partitions"]
            msas = dict((key, []) for key in partitions)
            for filename in files_dd["filenames"]:
                msa = MSA.from_file(filename)
                if self._excluded:
                    msa = msa.exclude(self._excluded)

                for (key, msa_part) in msa.split(partitions).iteritems():
                    msas[key].append(msa_part)

            msas.pop("X", None)
            for (key, msa_parts) in sorted(msas.iteritems()):
                merged_msa = MSA.join(*msa_parts)
                if self._reduce:
                    merged_msa = merged_msa.reduce()

                if merged_msa is not None:
                    merged_msas.append(("%s_%s" % (name, key),
                                        merged_msa))

        out_fname_phy = reroot_path(temp, self._out_prefix + ".phy")
        with open(out_fname_phy, "w") as output_phy:
            final_msa = MSA.join(*(msa for (_, msa) in merged_msas))
            output_phy.write(interleaved_phy(final_msa))

        partition_end = 0
        out_fname_parts = reroot_path(temp, self._out_prefix + ".partitions")
        with open(out_fname_parts, "w") as output_part:
            for (name, msa) in merged_msas:
                length = msa.seqlen()
                output_part.write("DNA, %s = %i-%i\n"
                                  % (name,
                                     partition_end + 1,
                                     partition_end + length))
                partition_end += length

    def _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._out_prefix + ".phy"),
                  self._out_prefix + ".phy")
        move_file(reroot_path(temp, self._out_prefix + ".partitions"),
                  self._out_prefix + ".partitions")




class FastaToPartitionsNode(Node):
    def __init__(self, infiles, out_partitions, partition_by = "123", dependencies = ()):
        if (len(partition_by) != 3):
            raise ValueError("Default 'partition_by' must be 3 entires long!")
        elif not isinstance(infiles, dict):
            raise TypeError("'infiles' must be a dictionary")
        elif any(len(dd.get("partition_by", "123")) != 3 for dd in infiles.itervalues()):
            raise ValueError("'partition_by' must be 3 entires long!")
        elif not all(isinstance(dd, dict) for dd in infiles.values()):
            raise TypeError("'infiles' must be a dictionary of dictionaries")
        elif not any(("name" in dd) for dd in infiles.values()):
            raise ValueError("'name' must be specified for all input files")
        elif any((set(dd) - _VALID_KEYS) for dd in infiles.values()):
            raise ValueError("Invalid keys found: %s" % ", ".join(set(dd) - _VALID_KEYS))

        self._infiles   = infiles
        self._out_part  = out_partitions
        self._part_by   = partition_by

        description  = "<FastaToPartitions (default: %s): %i file(s) -> '%s'>" % \
            (partition_by, len(infiles), out_partitions)

        Node.__init__(self,
                      description  = description,
                      input_files  = infiles.keys(),
                      output_files = out_partitions,
                      dependencies = dependencies)


    def _run(self, _config, temp):
        end = 0
        partitions = collections.defaultdict(list)
        for (filename, msa) in _read_sequences(self._infiles):
            length = msa.seqlen()
            start, end = end + 1, end + length

            for (group, offsets) in self._get_partition_by(filename):
                if len(offsets) != 3:
                    parts = [("%i-%i\\3" % (start + offset, end)) for offset in offsets]
                else:
                    parts = ["%i-%i" % (start, end)]

                name = "%s_%s" % (self._infiles[filename]["name"], group)
                partitions[name].extend(parts)

        with open(reroot_path(temp, self._out_part), "w") as part_file:
            for (name, parts) in sorted(partitions.items()):
                part_file.writelines("DNA, %s = %s\n" % (name, ", ".join(parts)))


    def  _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._out_part), self._out_part)


    def _get_partition_by(self, filename):
        groups = self._infiles[filename].get("partition_by", self._part_by)

        partition_by = {}
        for (group, offset) in zip(groups, range(3)):
            partition_by.setdefault(group, []).append(offset)

        return list(sorted(partition_by.items()))



class FastaToInterleavedPhyNode(Node):
    def __init__(self, infiles, out_phy, add_flag = False, dependencies = ()):
        self._add_flag  = add_flag
        self._out_phy   = out_phy

        description  = "<FastaToInterleavedPhy: %i file(s) -> '%s'%s>" % \
            (len(infiles), out_phy, (" (w/ flag)" if add_flag else ""))

        Node.__init__(self,
                      description  = description,
                      input_files  = infiles,
                      output_files = [out_phy],
                      dependencies = dependencies)


    def _run(self, _config, temp):
        msa = MSA.join(*(MSA.from_file(filename) for filename in sorted(self.input_files)))

        with open(reroot_path(temp, self._out_phy), "w") as output:
            output.write(interleaved_phy(msa, add_flag = self._add_flag))


    def  _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._out_phy), self._out_phy)



class FastaToSequentialPhyNode(Node):
    def __init__(self, infiles, out_phy, add_flag = False, dependencies = ()):
        self._add_flag  = add_flag
        self._out_phy   = out_phy

        description  = "<FastaToInterleavedPhy: %i file(s) -> '%s'%s>" % \
            (len(infiles), out_phy, (" (w/ flag)" if add_flag else ""))

        Node.__init__(self,
                      description  = description,
                      input_files  = infiles,
                      output_files = [out_phy],
                      dependencies = dependencies)


    def _run(self, _config, temp):
        # Read and check that MSAs share groups
        msas = [MSA.from_file(filename) for filename in sorted(self.input_files)]
        MSA.validate(*msas)

        blocks = []
        for msa in msas:
            blocks.append(sequential_phy(msa, add_flag = self._add_flag))

        with open(reroot_path(temp, self._out_phy), "w") as output:
            output.write("\n\n".join(blocks))


    def  _teardown(self, _config, temp):
        move_file(reroot_path(temp, self._out_phy), self._out_phy)





def _read_sequences(filenames):
    results = {}
    for filename in filenames:
        results[filename] = MSA.from_file(filename)
    MSA.validate(*results.values())

    return results.iteritems()



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from itertools import izip
from collections import defaultdict

from paleomix.common.sequences import split
from paleomix.common.fileutils import open_ro
from paleomix.common.formats.fasta import FASTA, FASTAError
from paleomix.common.sequences import NT_CODES, encode_genotype
from paleomix.common.utilities import safe_coerce_to_frozenset


class MSAError(FASTAError):
    pass


class MSA(frozenset):
    """Represents a Multiple Sequence Alignment of FASTA records."""

    def __new__(cls, sequences):
        records, names = [], set()
        for record in sequences:
            if record.name in names:
                raise MSAError("Duplicate name found in FASTA records: %r" % record.name)
            records.append(record)
            names.add(record.name)

        if not records:
            raise MSAError("MSA does not contain any sequences")

        instance = frozenset.__new__(cls, records)
        MSA.validate(instance)
        return instance

    def seqlen(self):
        """Retrurns the length of the sequences in the MSA."""
        return len(iter(self).next().sequence)

    def exclude(self, names):
        """Builds a new MSA that excludes the named set of records."""
        _, excluded, _ = self._group(names)
        return MSA(excluded)

    def select(self, names):
        """Builds a new MSA that includes only the named set of records."""
        included, _, _ = self._group(names)
        return MSA(included)

    def reduce(self):
        columns = []
        uncalled = frozenset("Nn-")
        for column in izip(*(record.sequence for record in self)):
            if (frozenset(column) - uncalled):
                columns.append(column)

        if not columns:
            return None

        records = []
        for (record, sequence) in izip(self, izip(*columns)):
            records.append(FASTA(record.name, record.meta, "".join(sequence)))

        return MSA(records)

    def filter_singletons(self, to_filter, filter_using):
        included, excluded, to_filter \
            = self._group(filter_using, to_filter)

        sequence = list(to_filter.sequence)
        sequences = [record.sequence.upper() for record in included]
        for (index, nts) in enumerate(zip(*sequences)):
            current_nt = sequence[index].upper()
            if current_nt in "N-":
                continue

            allowed_nts = set()
            for allowed_nt in nts:
                if allowed_nt not in "N-":
                    allowed_nts.update(NT_CODES[allowed_nt])
            filtered_nts = frozenset(NT_CODES[current_nt]) & allowed_nts

            if not filtered_nts:
                filtered_nts = "N"

            genotype = encode_genotype(filtered_nts)
            if genotype != current_nt:
                sequence[index] = genotype.lower()
        new_record = FASTA(to_filter.name,
                           to_filter.meta,
                           "".join(sequence))

        return MSA([new_record] + included + excluded)


    def split(self, split_by = "123"):
        """Splits a MSA and returns a dictionary of keys to MSAs,
        using the keys in the 'split_by' parameter at the top
        level. See also paleomix.common.sequences.split."""
        self.validate(self)
        if not split_by:
            raise TypeError("No partitions to split by specified")

        results = dict((key, set()) for key in split_by)
        for record in self:
            for (key, partition) in split(record.sequence, split_by).iteritems():
                results[key].add(FASTA(record.name, None, partition))

        for (key, value) in results.items():
            results[key] = MSA(value)

        return results

    @classmethod
    def join(cls, *msas):
        """Merge multiple MSAs into a single MSA, by concatenating sequences in
        the order of the passed MSAs. Sequences are joined by name, and all MSAs
        must therefore contain the same set of sequence names. Meta information
        is not preserved."""
        cls.validate(*msas)

        merged = defaultdict(list)
        for msa in msas:
            for record in msa:
                merged[record.name].append(record.sequence)

        sequences = []
        for (name, sequence) in merged.iteritems():
            sequences.append(FASTA(name, None, "".join(sequence)))
        return MSA(sequences)

    @classmethod
    def from_lines(cls, lines):
        """Parses a MSA from a file/list of lines, and returns a dictionary
        of names to sequences. If read_meta is True, meta information included
        after the first space in header of each sequence:
          >NAME META-INFORMATION
          SEQUENCE
        As suggested above, sequences are expected to be in FASTA format."""
        return MSA(FASTA.from_lines(lines))

    @classmethod
    def from_file(cls, filename):
        """Reads a MSA from the specified filename. The file may
        be uncompressed, gzipped or bzipped. See also 'MSA.from_lines'."""
        fasta_file = open_ro(filename)
        try:
            return MSA.from_lines(fasta_file)
        except MSAError, error:
            raise MSAError("%s in file %r" % (error, filename))
        finally:
            fasta_file.close()

    def to_file(self, fileobj):
        for fst in sorted(self):
            fileobj.write(str(fst))

    @classmethod
    def validate(cls, *msas):
        """Validates one or more MSAs, requiring:
        1. Identical sets of sequence names across all MSAs.
        2. That all names are non-empty strings.
        3. That all sequences are of the same length (per MSA).
        4. That no empty MSA (no sequences) are specified."""
        if not msas:
            raise TypeError("No MSAs given as arguments")

        seqs_all = msas[0].names()
        seqs_common = set(seqs_all)
        for msa in msas:
            if len(set(len(record.sequence) for record in msa)) != 1:
                raise MSAError("MSA contains sequences of differing lengths")

            seqs_all.update(msa.names())
            seqs_common &= set(msa.names())

        if seqs_all != seqs_common:
            raise MSAError("Some sequences not found in all MSAs: '%s'" \
                           % ("', '".join(seqs_all - seqs_common),))

    def __repr__(self):
        def _fasta_to_str(fst):
            return "FASTA(%r, %r, %r)" % \
              (fst.name, fst.meta, fst.sequence)
        return "MSA(%s)" % (", ".join(map(_fasta_to_str, sorted(self))))

    def names(self):
        return set(record.name for record in self)


    def _group(self, selection, extra = None):
        selection = safe_coerce_to_frozenset(selection)
        if (extra in selection):
            raise MSAError("Key used for multiple selections: %r" % extra)
        elif not selection:
            raise ValueError("No FASTA names given")

        missing_keys = selection - self.names()
        if missing_keys:
            raise KeyError("Key(s) not found: %r" % (", ".join(map(str, missing_keys))))

        included, excluded, other = [], [], None
        for record in self:
            if record.name in selection:
                included.append(record)
            elif record.name != extra:
                excluded.append(record)
            else:
                other = record

        return included, excluded, other



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys
import pysam

from paleomix.common.vcfwrap import \
     get_ml_genotype

import paleomix.common.timer as timer


def read_bed_records(filename):
    """Reads a bed-file (i.e. for a set of regions of interest), and returns
    a sorted list containing each line as a tuple containing the contig name,
    the start position, and the end position."""
    regions = []
    bed_parser = pysam.asBed()
    with open(filename) as bed_file:
        for line in bed_file:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            regions.append(bed_parser(line, len(line)))
    return regions


def select_vcf_records(bed_records, vcf_records):
    """Returns an iterable of VCF records, corresponding to the contents of each
    region specified by the BED records. Records are returned at most once, even
    if covered by multiple BED records."""
    contigs    = frozenset(vcf_records.contigs)
    vcf_parser = pysam.asVCF()

    # Timer class used processing progress; meant primarily for BAM files
    progress   = timer.BAMTimer(None)

    # Cache of positions observed for this contig, to prevent returning
    # positions in overlapping regions multiple times
    contig_cache = None
    contig_cache_name = None

    for bed in sorted(bed_records):
        if bed.contig not in contigs:
            # Skip contigs for which no calls have been made (e.g. due to
            # low coverage. Otherwise Pysam raises an exception.
            continue
        elif contig_cache_name != bed.contig:
            # Reset cache per contig, to save memory
            contig_cache = set()
            contig_cache_name = bed.contig

        for record in vcf_records.fetch(bed.contig, bed.start, bed.end, parser = vcf_parser):
            progress.increment()

            if record.pos in contig_cache:
                # We've already reported this VCF record
                continue

            contig_cache.add(record.pos)
            # Skip records filtered by VCF_filter
            if record.filter in ('.', "PASS"):
                yield record
    progress.finalize()


def main(argv):
    if len(argv) != 2:
        sys.stderr.write("Usage: %s <BED-file> <VCF.bgz>\n")
        return 1

    sites                 = 0
    sites_non_ref         = 0
    sites_homo_non_ref    = 0
    sites_het_one_non_ref = 0
    sites_het_two_non_ref = 0

    vcf_records = pysam.Tabixfile(argv[1])
    bed_records = read_bed_records(argv[0])

    for record in select_vcf_records(bed_records, vcf_records):
        if record.alt != '.':
            # Get the most likely diploid genotype
            nt_a, nt_b = get_ml_genotype(record)
            if (nt_a, nt_b) == ('N', 'N'):
                # Skip sites with no most likely genotype
                continue

            sites += 1
            sites_non_ref += 1
            if nt_a == nt_b:
                sites_homo_non_ref += 1
            elif record.ref not in (nt_a, nt_b):
                sites_het_two_non_ref += 1
            else:
                sites_het_one_non_ref += 1
        else:
            # Heterozygous for the reference allele
            sites += 1

    print
    print "%i sites kept after filtering:" % (sites,)
    print " % 10i homozygous sites containing the reference allele (%.2f%%)" % (sites - sites_non_ref, 100.0 * (sites - sites_non_ref) / float(sites))
    print " % 10i heterozygous sites containing the reference and a non-reference allele (%.2f%%)" % (sites_het_one_non_ref, (100.0 * sites_het_one_non_ref) / sites)
    print " % 10i homozygous sites containing a single non-reference allele (%.2f%%)" % (sites_homo_non_ref, (100.0 * sites_homo_non_ref) / sites)
    print " % 10i heterozygous sites containing two different non-reference alleles (%.2f%%)" % (sites_het_two_non_ref, (100.0 * sites_het_two_non_ref) / sites)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python -3
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import sys
import math
import gzip
import random

from optparse import \
    OptionParser, \
    OptionGroup

from paleomix.common.sequences import \
    reverse_complement
from paleomix.common.formats.fasta import \
    FASTA
from paleomix.common.utilities import \
    fragment
from paleomix.common.sampling import \
    weighted_sampling


def _dexp(lambda_value, position):
    return lambda_value * math.exp(-lambda_value * position)


def _rexp(lambda_value, rng):
    return - math.log(rng.random()) / lambda_value


def toint(value):
    return int(round(value))


# Adapter added to the 5' end of the forward strand (read from 5' ...)
PCR1 = "AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC%sATCTCGTATGCCGTCTTCTGCTTG"
# Adapter added to the 5' end of the reverse strand (read from 3' ...):
# rev. compl of the forward
PCR2 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT"


def _get_indel_length(indel_lambda, rng):
    return 1 + toint(_rexp(indel_lambda, rng))


def _get_weighted_choices(rng, sub_rate, indel_rate):
    choices_by_nt = {}
    for src_nt in "ACGT":
        choices = "ACGTID"
        probs = [sub_rate / 4] * 4     # ACGT
        probs += [indel_rate / 2] * 2  # ID
        probs[choices.index(src_nt)] = 1 - sum(probs) + sub_rate / 4
        choices_by_nt[src_nt] = weighted_sampling(choices, probs, rng)
    return choices_by_nt


def _mutate_sequence(rng, choices, refseq, indel_lambda=0):
    position = 0
    sequence, positions = [], []
    while position < len(refseq):
        ref_nt = refseq[position]
        if ref_nt not in "ACGT":
            read_nt = rng.choice("ACGT")
        else:
            read_nt = choices[ref_nt].next()

        if read_nt == "D":
            for _ in xrange(_get_indel_length(indel_lambda, rng)):
                position += 1
        elif read_nt == "I":
            for _ in xrange(_get_indel_length(indel_lambda, rng)):
                sequence.append(rng.choice("ACGT"))
                positions.append(position)
        else:
            sequence.append(read_nt)
            positions.append(position)
            position += 1
    return "".join(sequence), positions


class Specimen(object):
    """Represents a specimen, from which samples are derived.

    These are mutated by the addition of changes to the sequence

    """
    def __init__(self, options, filename):
        genome = list(FASTA.from_file(filename))
        assert len(genome) == 1, len(genome)

        self._genome = genome[0].sequence.upper()
        self._sequence = None
        self._positions = None
        self._annotations = None

        self._mutate(options)

    def _mutate(self, options):
        rng = random.Random(options.specimen_seed)
        choices = _get_weighted_choices(rng, options.specimen_sub_rate,
                                        options.specimen_indel_rate)
        self._sequence, self._positions = \
            _mutate_sequence(rng, choices, self._genome,
                             options.specimen_indel_lambda)

    @property
    def sequence(self):
        return self._sequence

    @property
    def positions(self):
        return self._positions

    @property
    def annotations(self):
        return self._annotations


class Sample(object):
    def __init__(self, options, specimen):
        self._specimen = specimen
        self._random = random.Random(options.sample_seed)
        self._options = options

        frac_endog = self._random.gauss(options.sample_endog_mu,
                                        options.sample_endog_sigma)
        self._frac_endog = min(1, max(0.01, frac_endog))
        self._endog_id = 0
        self._contam_id = 0

    def get_fragment(self):
        """Returns either a DNA fragmnet, representing either a fragment of
        the sample genome, or a randomly generated DNA sequence representing
        contaminant DNA that is not related to the species."""
        if self._random.random() <= self._frac_endog:
            return self._get_endogenous_sequence()
        return self._get_contaminant_sequence()

    def _get_contaminant_sequence(self):
        length = self._get_frag_len()
        sequence = [self._random.choice("ACGT") for _ in xrange(length)]

        self._contam_id += 1
        name = "Seq_junk_%i" % (self._contam_id,)
        return (False, name, "".join(sequence))

    def _get_endogenous_sequence(self):
        length = self._get_frag_len()
        max_position = len(self._specimen.sequence) - length
        position = self._random.randint(0, max_position)
        strand = self._random.choice(("fw", "rv"))

        sequence = self._specimen.sequence[position:position + length]
        real_pos = self._specimen.positions[position]
        if strand == "rv":
            sequence = reverse_complement("".join(sequence))

        self._endog_id += 1
        name = "Seq_%i_%i_%i_%s" % (self._endog_id, real_pos, length, strand)
        return (True, name, sequence)

    def _get_frag_len(self):
        length = toint(self._random.gauss(self._options.sample_frag_len_mu,
                                          self._options.sample_frag_len_sigma))

        return max(self._options.sample_frag_len_min,
                   min(self._options.sample_frag_len_max, length))


class Damage(object):
    def __init__(self, options, sample):
        self._options = options
        self._sample = sample
        self._random = random.Random(options.damage_seed)
        self._rates = self._calc_damage_rates(options)

    def get_fragment(self):
        is_endogenous, name, sequence = self._sample.get_fragment()
        if is_endogenous and self._options.damage:
            sequence = self._damage_sequence(sequence)
        return (name, sequence)

    def _damage_sequence(self, sequence):
        result = []
        length = len(sequence)
        for (position, nucleotide) in enumerate(sequence):
            if nucleotide == "C":
                if self._random.random() < self._rates[position]:
                    nucleotide = "T"
            elif nucleotide == "G":
                rv_position = length - position - 1
                if self._random.random() < self._rates[rv_position]:
                    nucleotide = "A"
            result.append(nucleotide)
        return "".join(result)

    @classmethod
    def _calc_damage_rates(cls, options):
        rate = options.damage_lambda
        rates = [_dexp(rate, position)
                 for position in range(options.sample_frag_len_max)]
        return rates


class Library(object):
    def __init__(self, options, sample):
        self._options = options
        self._sample = sample
        self._cache = []
        self._rng = random.Random(options.library_seed)

        self.barcode = options.library_barcode
        if self.barcode is None:
            self.barcode = "".join(self._rng.choice("ACGT") for _ in range(6))
        assert len(self.barcode) == 6, options.barcode

        pcr1 = PCR1 % (self.barcode,)
        self.lanes = self._generate_lanes(options, self._rng, sample, pcr1)

    @classmethod
    def _generate_lanes(cls, options, rng, sample, pcr1):
        lane_counts = []
        for _ in xrange(options.lanes_num):
            lane_counts.append(toint(random.gauss(options.lanes_reads_mu,
                                                  options.lanes_reads_sigma)))
        reads = cls._generate_reads(options, rng, sample,
                                    sum(lane_counts), pcr1)

        lanes = []
        for count in lane_counts:
            lanes.append(Lane(options, reads[:count]))
            reads = reads[count:]
        return lanes

    @classmethod
    def _generate_reads(cls, options, rng, sample, minimum, pcr1):
        reads = []
        while len(reads) < minimum:
            name, sequence = sample.get_fragment()
            cur_forward = sequence + pcr1
            cur_reverse = reverse_complement(sequence) + PCR2
            # Number of PCR copies -- minimum 1
            num_dupes = toint(_rexp(options.library_pcr_lambda, rng)) + 1
            for dupe_id in xrange(num_dupes):
                cur_name = "%s_%s" % (name, dupe_id)
                reads.append((cur_name, cur_forward, cur_reverse))
        random.shuffle(reads)
        return reads


class Lane(object):
    def __init__(self, options, reads):
        rng = random.Random()
        choices = _get_weighted_choices(rng, options.reads_sub_rate,
                                        options.reads_indel_rate)

        self._sequences = []
        for (name, forward, reverse) in reads:
            forward, _ = _mutate_sequence(rng, choices, forward,
                                          options.reads_indel_lambda)

            if len(forward) < options.reads_len:
                forward += "A" * (options.reads_len - len(forward))
            elif len(forward) > options.reads_len:
                forward = forward[:options.reads_len]

            reverse, _ = _mutate_sequence(rng, choices, reverse,
                                          options.reads_indel_lambda)

            if len(reverse) < options.reads_len:
                reverse += "T" * (options.reads_len - len(reverse))
            elif len(reverse) > options.reads_len:
                reverse = reverse[:options.reads_len]

            self._sequences.append((name, "".join(forward), "".join(reverse)))

    @property
    def sequences(self):
        return self._sequences


def parse_args(argv):
    parser = OptionParser()

    group = OptionGroup(parser, "Specimen")
    group.add_option("--specimen-seed", default=None,
                     help="Seed used to initialize the 'speciment', for the "
                          "creation of a random genotype. Set to a specific "
                          "values if runs are to be done for the same "
                          "genotype.")
    group.add_option("--specimen-sub-rate", default=0.005,  type=float)
    group.add_option("--specimen-indel-rate", default=0.0005, type=float)
    group.add_option("--specimen-indel-lambda", default=0.9, type=float)
    parser.add_option_group(group)

    group = OptionGroup(parser, "Samples from specimens")
    group.add_option("--sample-seed", default=None)
    group.add_option("--sample-frag-length-mu",
                     dest="sample_frag_len_mu", default=100, type=int)
    group.add_option("--sample-frag-length-sigma",
                     dest="sample_frag_len_sigma", default=30, type=int)
    group.add_option("--sample-frag-length-min",
                     dest="sample_frag_len_min", default=0, type=int)
    group.add_option("--sample-frag-length-max",
                     dest="sample_frag_len_max", default=500, type=int)
    group.add_option("--sample-endogenous_mu",
                     dest="sample_endog_mu", default=0.75, type=float)
    group.add_option("--sample-endogenous_sigma",
                     dest="sample_endog_sigma", default=0.10, type=float)
    parser.add_option_group(group)

    group = OptionGroup(parser, "Post mortem damage of samples")
    group.add_option("--damage", dest="damage",
                     default=False, action="store_true")
    group.add_option("--damage-seed", dest="damage_seed", default=None)
    group.add_option("--damage-lambda", dest="damage_lambda",
                     default=0.25, type=float)
    parser.add_option_group(group)

    group = OptionGroup(parser, "Libraries from samples")
    group.add_option("--library-seed", dest="library_seed", default=None)
    group.add_option("--library-pcr-lambda", dest="library_pcr_lambda",
                     default=3, type=float)
    group.add_option("--library-barcode", dest="library_barcode", default=None)
    parser.add_option_group(group)

    group = OptionGroup(parser, "Lanes from libraries")
    group.add_option("--lanes", dest="lanes_num", default=3, type=int)
    group.add_option("--lanes-reads-mu", dest="lanes_reads_mu",
                     default=10000, type=int)
    group.add_option("--lanes-reads-sigma", dest="lanes_reads_sigma",
                     default=2500, type=int)
    group.add_option("--lanes-reads-per-file", dest="lanes_per_file",
                     default=2500, type=int)
    parser.add_option_group(group)

    group = OptionGroup(parser, "Reads from lanes")
    group.add_option("--reads-sub-rate", dest="reads_sub_rate",
                     default=0.005, type=float)
    group.add_option("--reads-indel-rate", dest="reads_indel_rate",
                     default=0.0005, type=float)
    group.add_option("--reads-indel-lambda",
                     dest="reads_indel_lambda", default=0.9, type=float)
    group.add_option("--reads-length", dest="reads_len", default=100, type=int)
    parser.add_option_group(group)

    options, args = parser.parse_args(argv)
    if len(args) != 2:
        sys.stderr.write("Usage: %s <genome> <prefix>\n" % sys.argv[0])
        return None, None
    return options, args


def main(argv):
    options, args = parse_args(argv)
    if not options:
        return 1

    print("Generating %i lane(s) of synthetic reads ...\nDISCLAIMER: For "
          "demonstration of PALEOMIX usage only; not for serious usage!"
          % (options.lanes_num,))

    specimen = Specimen(options, args[0])
    sample = Sample(options, specimen)
    damage = Damage(options, sample)
    library = Library(options, damage)

    for (lnum, lane) in enumerate(library.lanes, start=1):
        fragments = fragment(options.lanes_per_file, lane.sequences)
        for (readsnum, reads) in enumerate(fragments, start=1):
            templ = "%s%s_L%i_R%%s_%02i.fastq.gz" % (args[1], library.barcode,
                                                     lnum, readsnum)

            print("  Writing %s" % (templ % "{Pair}",))
            with gzip.open(templ % 1, "w") as out_1:
                with gzip.open(templ % 2, "w") as out_2:
                    for (name, seq_1, seq_2) in reads:
                        out_1.write("@%s%s/1\n%s\n" % (library.barcode, name, seq_1))
                        out_1.write("+\n%s\n" % ("I" * len(seq_1),))
                        out_2.write("@%s%s/2\n%s\n" % (library.barcode, name, seq_2))
                        out_2.write("+\n%s\n" % ("H" * len(seq_2),))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys
import datetime
import itertools
import collections

from paleomix.common.timer import \
    BAMTimer
from paleomix.common.text import \
    padded_table
from paleomix.common.bamfiles import \
    BAMRegionsIter

from paleomix.tools.bam_stats.common import \
    collect_references, \
    collect_readgroups, \
    main_wrapper


##############################################################################
##############################################################################
##

# Maximum depth to record, and hence the number of columns in output table
_MAX_DEPTH = 200
# Maximum number of count patterns (numbers of bases per library for a given
# site) to cache for bulk processing; see MappingsToTotals for implementation
_MAX_CACHE_SIZE = 10000


# Header prepended to output tables
_HEADER = """# Timestamp: %s
#
# Columns:
#   Contig:   Contig, chromosome, or feature for which a depth histogram was
#             created. Unnamed features are named after the chromosome or
#             contig on which they are located, with a star appended. For
#             example "chr1*". If the maximum number of contigs was exceeded,
#             these are collapsed into one meta-contig named "<Genome>".
#   Size:     The total size of the region. Multiple features with the same
#             name are combined into one row, with the size representing the
#             total for these. Note that overlapping bases are counted 2 (or
#             more) times.
#   MaxDepth: Maximum depth to use when calling SNPs, in order to exclude
#             (at least) the 0.5%% most extreme sites based on read depth,
#             not including sites with depth 0.
#   MD_*:     Fraction of sites with a minimum depth of 1-200.
#"""


##############################################################################
##############################################################################

class MappingToTotals(object):
    def __init__(self, totals, region, smlbid_to_smlb):
        self._region = region
        self._totals = totals
        self._map_by_smlbid, self._totals_src_and_dst \
            = self._build_mappings(totals, region.name, smlbid_to_smlb)
        self._cache = collections.defaultdict(int)

    def process_counts(self, counts, last_pos, cur_pos):
        start = self._region.start
        end = self._region.end

        # Pileups tends to contain identical stretches, so
        # try to avoid repeated lookups by aggregating these
        repeats = 1
        last_count = None
        while counts and (last_pos < cur_pos):
            count = counts.popleft()
            if start <= last_pos < end:
                if count == last_count:
                    repeats += 1
                else:
                    if last_count is not None:
                        self._cache[tuple(last_count)] += repeats
                    last_count = count
                    repeats = 1
            last_pos += 1

        if last_count is not None:
            self._cache[tuple(last_count)] += repeats

        if len(self._cache) > _MAX_CACHE_SIZE:
            self.finalize()

    def finalize(self):
        """Process cached counts."""
        for (count, multiplier) in self._cache.iteritems():
            self._update_totals(count, multiplier)
        self._cache.clear()

    def _update_totals(self, count, multiplier=1):
        for (smlbid, count) in enumerate(count):
            if count:
                for lst in self._map_by_smlbid[smlbid]:
                    lst[0] += count

        for (dst_counts, src_count) in self._totals_src_and_dst:
            if src_count[0]:
                dst_counts[src_count[0]] += multiplier
                src_count[0] = 0

    @classmethod
    def _build_mappings(cls, totals, name, smlbid_to_smlb):
        # Accumulators mapped by sample+library IDs
        totals_by_smlbid = [None] * len(smlbid_to_smlb)
        # Accumulators mapped by the corresponding table keys
        totals_by_table_key = {}

        for (smlbid, (sm_key, lb_key)) in enumerate(smlbid_to_smlb):
            keys = [('*', '*', '*'),
                    (sm_key, '*', '*'),
                    (sm_key, '*', name),
                    (sm_key, lb_key, '*'),
                    (sm_key, lb_key, name)]

            mappings = cls._nonoverlapping_mappings(keys, totals,
                                                    totals_by_table_key)
            totals_by_smlbid[smlbid] = mappings

        totals_src_and_dst = []
        for (key, dst) in totals_by_table_key.iteritems():
            totals_src_and_dst.append((totals[key], dst))

        return totals_by_smlbid, totals_src_and_dst

    @classmethod
    def _nonoverlapping_mappings(cls, keys, totals, totals_by_table_key):
        """Returns a tuple of accumulators for a given set of table keys. As
        multiple table keys may share the same accumulator (e.g. if there is
        only one sample, then sample "*" and that sample will be identical),
        the tuple of accumulators may contain fewer items than keys."""

        mapping = []
        totals_used = set()
        for key in keys:
            # Check that accumulator is not already included
            totals_id = id(totals[key])
            if totals_id not in totals_used:
                totals_used.add(totals_id)
                accumulator = totals_by_table_key.setdefault(key, [0])
                mapping.append(accumulator)
        return tuple(mapping)


##############################################################################
##############################################################################

def calc_max_depth(counts):
    counts = dict(counts)
    counts.pop(0, None)

    running_total = sum(counts.values())
    if not running_total:
        return "NA"

    total = float(running_total)
    for (index, count) in sorted(counts.items()):
        # Stop when less than the 0.5% most extreme values are included
        if running_total / total < 0.005:
            # The max is inclusive, so return the depth just before this one
            return index - 1
        running_total -= count

    return "NA"


def print_table(handle, args, totals):
    lengths = collect_references(args, handle)

    if args.outfile == "-":
        output_handle = sys.stdout
    else:
        output_handle = open(args.outfile, "w")

    with output_handle:
        rows = build_table(args.target_name, totals, lengths)
        output_handle.write(_HEADER % datetime.datetime.now().isoformat())
        output_handle.write("\n")
        for line in padded_table(rows):
            output_handle.write(line)
            output_handle.write("\n")


def calculate_depth_pc(counts, length):
    final_counts = [0] * (_MAX_DEPTH + 1)
    for (depth, count) in counts.iteritems():
        final_counts[min(_MAX_DEPTH, depth)] += count

    running_total = sum(final_counts)
    total = float(length)
    for count in final_counts[1:]:
        yield "%.4f" % (running_total / total,)
        running_total -= count


def build_table(name, totals, lengths):
    header = ["Name", "Sample", "Library", "Contig", "Size", "MaxDepth"]
    for index in xrange(1, _MAX_DEPTH + 1):
        header.append("MD_%03i" % (index,))

    rows = [header]
    last_sm = last_lb = None
    for ((sm_key, lb_key, ct_key), counts) in sorted(totals.items()):
        if (sm_key != last_sm) and (last_sm is not None):
            rows.extend("##")
        elif (lb_key != last_lb) and (last_lb is not None):
            rows.append("#")
        last_sm, last_lb = sm_key, lb_key

        if ct_key == "*":
            length = sum(lengths.itervalues())
        else:
            length = lengths[ct_key]

        row = [name, sm_key, lb_key, ct_key, str(length),
               str(calc_max_depth(counts))]
        row.extend(calculate_depth_pc(counts, length))

        rows.append(row)

    return rows


##############################################################################
##############################################################################

def build_key_struct(args, handle):
    structure = collections.defaultdict(set)
    for readgroup in collect_readgroups(args, handle).itervalues():
        lb_key = readgroup["LB"]
        sm_key = readgroup["SM"]
        structure[sm_key].add(lb_key)

    return structure


def build_new_dicts(totals, dst_sm, dst_lb, references):
    totals[(dst_sm, dst_lb, '*')] = collections.defaultdict(int)
    for contig in references:
        totals[(dst_sm, dst_lb, contig)] = collections.defaultdict(int)


def reuse_dicts(totals, dst_sm, dst_lb, src_sm, src_lb, references):
    totals[(dst_sm, dst_lb, '*')] = totals[(src_sm, src_lb, '*')]
    for contig in references:
        totals[(dst_sm, dst_lb, contig)] = totals[(src_sm, src_lb, contig)]


def build_totals_dict(args, handle):
    references = tuple(collect_references(args, handle))
    structure = build_key_struct(args, handle)

    totals = {}
    for (sm_key, libraries) in structure.iteritems():
        for lb_key in libraries:
            if len(references) == 1:
                key = references[0]
                counts = collections.defaultdict(int)
                totals[(sm_key, lb_key, key)] = counts
                totals[(sm_key, lb_key, '*')] = counts
            else:
                build_new_dicts(totals, sm_key, lb_key, references)

        if len(libraries) == 1:
            key = list(libraries)[0]
            reuse_dicts(totals, sm_key, '*', sm_key, key, references)
        else:
            build_new_dicts(totals, sm_key, '*', references)

    if len(structure) == 1:
        key = list(structure)[0]
        reuse_dicts(totals, '*', '*', key, '*', references)
    else:
        build_new_dicts(totals, '*', '*', references)

    return totals


def count_bases(args, counts, record, rg_to_smlbid, template):
    for _ in xrange(record.alen - len(counts)):
        counts.append(list(template))

    key = rg_to_smlbid[args.get_readgroup_func(record)]
    index = 0
    for (cigar, count) in record.cigar:
        if cigar in (0, 7, 8):
            for counter in itertools.islice(counts, index, index + count):
                counter[key] += 1
            index += count
        elif cigar in (2, 3, 6):
            index += count


def build_rg_to_smlbid_keys(args, handle):
    """Returns a dictionary which maps a readgroup ID to an index value,
    as well as a list containing a tuple (samples, library) corresponding
    to each index. Typically, this list will be shorter than the map of read-
    groups, as multiple read-groups will map to the same sample / library.
    """

    rg_to_lbsmid = {}
    lbsm_to_lbsmid = {}
    lbsmid_to_smlb = []
    for (key_rg, readgroup) in collect_readgroups(args, handle).iteritems():
        key_sm = readgroup["SM"]
        key_lb = readgroup["LB"]

        key_lbsm = (key_sm, key_lb)
        if key_lbsm not in lbsm_to_lbsmid:
            lbsm_to_lbsmid[key_lbsm] = len(lbsm_to_lbsmid)
            lbsmid_to_smlb.append(key_lbsm)

        rg_to_lbsmid[key_rg] = lbsm_to_lbsmid[key_lbsm]
    return rg_to_lbsmid, lbsmid_to_smlb


def process_file(handle, args):
    timer = BAMTimer(handle, step=1000000)

    last_tid = 0
    totals = build_totals_dict(args, handle)
    rg_to_smlbid, smlbid_to_smlb = build_rg_to_smlbid_keys(args, handle)
    template = [0] * len(smlbid_to_smlb)

    for region in BAMRegionsIter(handle, args.regions):
        if region.name is None:
            # Trailing unmapped reads
            continue
        elif not args.regions and (handle.nreferences > args.max_contigs):
            region.name = '<Genome>'

        last_pos = 0
        counts = collections.deque()
        mapping = MappingToTotals(totals, region, smlbid_to_smlb)
        for (position, records) in region:
            mapping.process_counts(counts, last_pos, position)

            for record in records:
                timer.increment(read=record)
                count_bases(args, counts, record, rg_to_smlbid, template)

            if (region.tid, position) < (last_tid, last_pos):
                sys.stderr.write("ERROR: Input BAM file is unsorted\n")
                return 1

            last_pos = position
            last_tid = region.tid

        # Process columns in region after last read
        mapping.process_counts(counts, last_pos, float("inf"))
        mapping.finalize()
    timer.finalize()

    print_table(handle, args, totals)

    return 0


def main(argv):
    return main_wrapper(process_file, argv, ".depths")

##############################################################################
##############################################################################

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Disable warnings for wierd function names
# pylint: disable=C0103
# Disable warnings caused by flexmock setups ("X is assigned to nothing")
# pylint: disable=W0106

import nose
import nose.tools
from nose.tools import assert_equal, assert_raises
from paleomix.common.testing import with_temp_folder

from flexmock import flexmock

import paleomix.atomiccmd.pprint
from paleomix.atomiccmd.command import AtomicCmd, CmdError
from paleomix.atomiccmd.sets import ParallelCmds, SequentialCmds


###############################################################################
###############################################################################
# Properties with same expected behavior for both Parallel/SequentialCmds

def test_atomicsets__properties():
    def _do_test(cls):
        cmd_mock_1 = AtomicCmd(("true",),
                               CHECK_A=id,
                               EXEC_1="false",
                               IN_1="/foo/bar/in_1.file",
                               IN_2="/foo/bar/in_2.file",
                               OUT_1="/bar/foo/out",
                               TEMP_OUT_1="out.log",
                               AUX_A="/aux/fA",
                               AUX_B="/aux/fB")
        cmd_mock_2 = AtomicCmd(("false",),
                               CHECK_A=list,
                               EXEC_1="echo",
                               EXEC_2="java",
                               IN_1="/foo/bar/in.file",
                               OUT_1="out.txt")

        obj = cls([cmd_mock_1, cmd_mock_2])
        assert_equal(obj.executables, cmd_mock_1.executables | cmd_mock_2.executables)
        assert_equal(obj.requirements, cmd_mock_1.requirements | cmd_mock_2.requirements)
        assert_equal(obj.input_files, cmd_mock_1.input_files | cmd_mock_2.input_files)
        assert_equal(obj.output_files, cmd_mock_1.output_files | cmd_mock_2.output_files)
        assert_equal(obj.auxiliary_files, cmd_mock_1.auxiliary_files | cmd_mock_2.auxiliary_files)
        assert_equal(obj.expected_temp_files, frozenset(["out", "out.txt"]))
        assert_equal(obj.optional_temp_files, cmd_mock_1.optional_temp_files | cmd_mock_2.optional_temp_files)

    for cls in (ParallelCmds, SequentialCmds):
        yield _do_test, cls


# Ensure that commands in a set doesn't clobber eachothers OUT files
def test_atomicsets__no_clobbering():
    def _do_test_atomicsets__no_clobbering(cls, kwargs_1, kwargs_2):
        cmd_1 = AtomicCmd("true", **kwargs_1)
        cmd_2 = AtomicCmd("true", **kwargs_2)
        assert_raises(CmdError, cls, [cmd_1, cmd_2])

    for cls in (ParallelCmds, SequentialCmds):
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"OUT_B": "/bar/out.txt"}
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_B": "out.txt"}
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"OUT_STDOUT": "/bar/out.txt"}
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_STDOUT": "out.txt"}
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"OUT_STDERR": "/bar/out.txt"}
        yield _do_test_atomicsets__no_clobbering, cls, {"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_STDERR": "out.txt"}


###############################################################################
###############################################################################
# Functions with same expected behavior for both Parallel/SequentialCmds

def test_atomicsets__commit():
    def _do_test_atomicsets__commit(cls):
        mocks = []
        for _ in range(3):
            cmd_mock = flexmock(AtomicCmd(["ls"]))
            cmd_mock.should_receive('commit').with_args("xTMPx").once.ordered
            mocks.append(cmd_mock)

        cls(mocks).commit("xTMPx")

    yield _do_test_atomicsets__commit, ParallelCmds
    yield _do_test_atomicsets__commit, SequentialCmds


def test_atomicsets__stdout():
    @nose.tools.raises(CmdError)
    def _do_test_atomicsets__stdout(cls):
        cmds = cls([AtomicCmd("ls")])
        cmds.stdout

    yield _do_test_atomicsets__stdout, ParallelCmds
    yield _do_test_atomicsets__stdout, SequentialCmds


def test_atomicsets__terminate():
    def _do_test_atomicsets__terminate(cls):
        mocks = []
        for _ in reversed(range(3)):
            cmd_mock = flexmock(AtomicCmd("true"))
            cmd_mock.should_receive('terminate').with_args().once
            mocks.append(cmd_mock)
        cmds = cls(mocks)
        cmds.terminate()

    yield _do_test_atomicsets__terminate, ParallelCmds
    yield _do_test_atomicsets__terminate, SequentialCmds


def test_atomicsets__str__():
    def _do_test_atomicsets__str__(cls):
        cmds = cls([AtomicCmd("ls")])
        assert_equal(paleomix.atomiccmd.pprint.pformat(cmds), str(cmds))

    yield _do_test_atomicsets__str__, ParallelCmds
    yield _do_test_atomicsets__str__, SequentialCmds


def test_atomicsets__duplicate_cmds():
    def _do_test_atomicsets__duplicate_cmds(cls):
        cmd_1 = AtomicCmd("true")
        cmd_2 = AtomicCmd("false")
        assert_raises(ValueError, cls, [cmd_1, cmd_2, cmd_1])

    yield _do_test_atomicsets__duplicate_cmds, ParallelCmds
    yield _do_test_atomicsets__duplicate_cmds, SequentialCmds


###############################################################################
###############################################################################
# Parallel commands

def test_parallel_commands__run():
    mocks = []
    for _ in range(3):
        cmd_mock = flexmock(AtomicCmd(["ls"]))
        cmd_mock.should_receive('run').with_args("xTMPx").once
        mocks.append(cmd_mock)

    cmds = ParallelCmds(mocks)
    cmds.run("xTMPx")


def test_parallel_commands__ready_single():
    def _do_test_parallel_commands__ready_single(value):
        cmd_mock = flexmock(AtomicCmd(["ls"]))
        cmd_mock.should_receive('ready').and_return(value).at_least.once
        cmds = ParallelCmds([cmd_mock])
        assert_equal(cmds.ready(), value)

    yield _do_test_parallel_commands__ready_single, True
    yield _do_test_parallel_commands__ready_single, False


def test_parallel_commands__ready_two():
    def _do_test_parallel_commands__ready_two(first, second, result):
        cmd_mock_1 = flexmock(AtomicCmd(["ls"]))
        cmd_mock_1.should_receive('ready').and_return(first).at_least.once
        cmd_mock_2 = flexmock(AtomicCmd(["ls"]))
        cmd_mock_2.should_receive('ready').and_return(second)
        cmds = ParallelCmds([cmd_mock_1, cmd_mock_2])
        assert_equal(cmds.ready(), result)

    yield _do_test_parallel_commands__ready_two, True,   True,  True
    yield _do_test_parallel_commands__ready_two, False,  True, False
    yield _do_test_parallel_commands__ready_two, True,  False, False
    yield _do_test_parallel_commands__ready_two, False, False, False


def test_parallel_commands__join_before_run():
    mocks = []
    for value in reversed(range(3)):
        cmd_mock = flexmock(AtomicCmd("true"))
        cmd_mock.should_receive('join').and_return([value]).never
        mocks.append(cmd_mock)
    cmds = ParallelCmds(mocks)
    assert_equal(cmds.join(), [None, None, None])


@with_temp_folder
def test_parallel_commands__join_after_run(temp_folder):
    cmds = ParallelCmds([AtomicCmd("true") for _ in range(3)])
    cmds.run(temp_folder)
    assert_equal(cmds.join(), [0, 0, 0])


def _setup_mocks_for_failure(*do_mocks):
    results = []
    for do_mock in do_mocks:
        if do_mock:
            mock = flexmock(AtomicCmd(("sleep", 10)))
            mock.should_receive('terminate')
            mock.should_receive('join').and_return(['SIGTERM'])
        else:
            mock = AtomicCmd("false")
        results.append(mock)
    return results


@with_temp_folder
def test_parallel_commands__join_failure_1(temp_folder):
    mocks = _setup_mocks_for_failure(False, True, True)
    cmds = ParallelCmds(mocks)
    cmds.run(temp_folder)
    assert_equal(cmds.join(), [1, 'SIGTERM', 'SIGTERM'])


@with_temp_folder
def test_parallel_commands__join_failure_2(temp_folder):
    mocks = _setup_mocks_for_failure(True, False, True)
    cmds = ParallelCmds(mocks)
    cmds.run(temp_folder)
    assert_equal(cmds.join(), ['SIGTERM', 1, 'SIGTERM'])


@with_temp_folder
def test_parallel_commands__join_failure_3(temp_folder):
    mocks = _setup_mocks_for_failure(True, True, False)
    cmds = ParallelCmds(mocks)
    cmds.run(temp_folder)
    assert_equal(cmds.join(), ['SIGTERM', 'SIGTERM', 1])


def test_parallel_commands__reject_sequential():
    command = AtomicCmd(["ls"])
    seqcmd = SequentialCmds([command])
    assert_raises(CmdError, ParallelCmds, [seqcmd])


def test_parallel_commands__accept_parallel():
    command = AtomicCmd(["ls"])
    parcmd = ParallelCmds([command])
    ParallelCmds([parcmd])


@nose.tools.raises(CmdError)
def test_parallel_commands__reject_noncommand():
    ParallelCmds([object()])


@nose.tools.raises(CmdError)
def test_parallel_commands__reject_empty_commandset():
    ParallelCmds([])


###############################################################################
###############################################################################
# Sequential commands

def test_sequential_commands__atomiccmds():
    mocks = []
    for _ in range(3):
        cmd_mock = flexmock(AtomicCmd(["ls"]))
        cmd_mock.should_receive('run').with_args("xTMPx").once
        cmd_mock.should_receive('join').with_args().and_return([0]).twice
        mocks.append(cmd_mock)

    cmds = SequentialCmds(mocks)
    assert not cmds.ready()
    cmds.run("xTMPx")
    assert cmds.ready()
    assert_equal(cmds.join(), [0, 0, 0])


@with_temp_folder
@nose.tools.timed(1)
def test_sequential_commands__abort_on_error_1(temp_folder):
    cmd_1 = AtomicCmd("false")
    cmd_2 = AtomicCmd(("sleep", 10))
    cmd_3 = AtomicCmd(("sleep", 10))
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(temp_folder)
    assert_equal(cmds.join(), [1, None, None])


@with_temp_folder
@nose.tools.timed(1)
def test_sequential_commands__abort_on_error_2(temp_folder):
    cmd_1 = AtomicCmd("true")
    cmd_2 = AtomicCmd("false")
    cmd_3 = AtomicCmd(("sleep", 10))
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(temp_folder)
    assert_equal(cmds.join(), [0, 1, None])


@with_temp_folder
@nose.tools.timed(1)
def test_sequential_commands__abort_on_error_3(temp_folder):
    cmd_1 = AtomicCmd("true")
    cmd_2 = AtomicCmd("true")
    cmd_3 = AtomicCmd("false")
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(temp_folder)
    assert_equal(cmds.join(), [0, 0, 1])


def test_sequential_commands__accept_parallel():
    command = AtomicCmd(["ls"])
    parcmd = ParallelCmds([command])
    SequentialCmds([parcmd])


def test_sequential_commands__accept_sequential():
    command = AtomicCmd(["ls"])
    seqcmd = SequentialCmds([command])
    SequentialCmds([seqcmd])


@nose.tools.raises(CmdError)
def test_sequential_commands__reject_noncommand():
    SequentialCmds([object()])


@nose.tools.raises(CmdError)
def test_sequential_commands__reject_empty_commandset():
    SequentialCmds([])



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import sys
import errno
import optparse
import logging

import paleomix.common.console as _cli


def initialize(config, template = None):
    """Takes an OptionParser object for which 'add_optiongroup' has
    been called, as well as a filename template (containing one '%i'
    field), and initializes logging for a PALEOMIX pipeline.

    If --log-file has not been specified, the template is used to
    create a new logfile in --temp-root, skipping existing logfiles
    by incrementing the counter value. If a --log-file has been
    specified, this file is always created / opened.

    If neither --log-file nor 'template' has been specified, then
    logging is only carried out by printing messages to STDERR."""
    global _INITIALIZED # pylint: disable=W0603
    if _INITIALIZED:
        raise RuntimeError("Attempting to initialize logging more than once")

    root  = logging.getLogger()
    root.setLevel(logging.INFO)
    root.addHandler(_PrintToConsole(logging.INFO))

    if config.log_file or template:
        # Verify that the template is functional up front
        template % (1,) # pylint: disable=W0104

        level = _LOGLEVELS[config.log_level]
        if config.log_file:
            handler = logging.FileHandler(config.log_file)
        else:
            handler = _LazyLogfile(config.temp_root, template)
        fmt = "%s\n%%(asctime)s -- %%(levelname)s:\n%%(message)s"
        formatter = logging.Formatter(fmt % ("-" * 60,))
        handler.setFormatter(formatter)
        handler.setLevel(level)
        root.addHandler(handler)

    _INITIALIZED = True


def add_optiongroup(parser, default = "warning"):
    """Adds an option-group to an OptionParser object, with options
    pertaining to logging. Note that 'initialize' expects the config
    object to have these options."""
    group  = optparse.OptionGroup(parser, "Logging")
    group.add_option("--log-file", default = None,
                     help = "Create the specified log-file and write any "
                            "messages to this file. By default, a log-file "
                            "will be generated in the folder specified using "
                            "--temp-root, but only when messages are logged")
    group.add_option("--log-level", default = default, type = "choice",
                     choices = ("info", "warning", "error", "debug"),
                     help = "Log messages to log-file at and above the "
                            "specified level; one of 'info', 'warning', "
                            "'error', or 'debug' [%default]")
    parser.add_option_group(group)


def get_logfile():
    return _LOGFILE


class _PrintToConsole(logging.Handler):
    """Logger that prints messages to the console using the
    paleomix.ui functions for colored text. Colors are blue
    for DEBUG, green for INFO (and unknown levels), yellow
    for WARNING, and red for ERROR and CRITICAL."""
    def __init__(self, level = logging.NOTSET):
        logging.Handler.__init__(self, level)

    def emit(self, record):
        func = self.get_ui_function(record.levelno)
        func(record.getMessage(), file = sys.stderr)

    @classmethod
    def get_ui_function(cls, level):
        if level in (logging.ERROR, logging.CRITICAL):
            return _cli.print_err
        elif level == logging.WARNING:
            return _cli.print_warn
        elif level == logging.DEBUG:
            return _cli.print_debug
        return _cli.print_info


class _LazyLogfile(logging.Handler):
    def __init__(self, folder, template):
        logging.Handler.__init__(self)
        self._folder   = folder
        self._template = template
        self._stream    = None
        self._handler   = None
        self._formatter = None


    def emit(self, record):
        if not self._handler:
            global _LOGFILE # pylint: disable = W0603
            _LOGFILE, self._stream = \
              _open_logfile(self._folder, self._template)
            self._handler = logging.StreamHandler(self._stream)
            self._handler.setFormatter(self._formatter)
        self._handler.emit(record)


    def flush(self):
        if self._handler:
            self._handler.flush()


    def setFormatter(self, form):
        logging.Handler.setFormatter(self, form)
        self._formatter = form


    def close(self):
        if self._handler:
            self._handler.close()
            self._stream.close()
            self._handler = None
            self._stream  = None


def _open_logfile(folder, template, start = 0):
    """Try to open a new logfile, taking steps to ensure that
    existing logfiles using the same template are not clobbered."""
    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
    while True:
        filename = os.path.join(folder, template % (start,))
        try:
            if not os.path.exists(filename):
                return filename, os.fdopen(os.open(filename, flags), "w")
        except OSError, error:
            if error.errno != errno.EEXIST:
                raise
        start += 1


_INITIALIZED = False
_LOGFILE     = None
_LOGLEVELS   = {
    'info'    : logging.INFO,
    'warning' : logging.WARNING,
    'error'   : logging.ERROR,
    'debug'   : logging.DEBUG,
}



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.node import CommandNode
from paleomix.atomiccmd.command import AtomicCmd

import paleomix.common.versions as versions


BEDTOOLS_VERSION = versions.Requirement(call   = ("bedtools", "--version"),
                                        search = r"bedtools v?(\d+)\.(\d+)\.(\d+)",
                                        checks = versions.GE(2, 15, 0))




class SlopBedNode(CommandNode):
    def __init__(self, infile, outfile, genome, from_start = 0, from_end = 0, strand_relative = False, dependencies = ()):
        if type(from_start) != type(from_end):
            raise ValueError("'from_start' and 'from_end' should be of same type!")

        call = ["bedtools", "slop",
                "-i", "%(IN_FILE)s",
                "-g", "%(IN_GENOME)s",
                "-l", str(from_start),
                "-r", str(from_end)]

        if strand_relative:
            call.append("-s")
        if type(from_start) is float:
            call.append("-pct")

        command = AtomicCmd(call,
                            IN_FILE    = infile,
                            IN_GENOME  = genome,
                            OUT_STDOUT = outfile,
                            CHECK_VERSION = BEDTOOLS_VERSION)

        description = "<SlopBed: '%s' -> '%s'>" % (infile, outfile)

        CommandNode.__init__(self,
                             description  = description,
                             command      = command,
                             dependencies = dependencies)
